
// Tm tiedosto sislt Datanhallintaluokan toteutuksen.

package datanhallinta;

import java.io.*;
 
/**Datanhallintaluokasta voi olla vain yksi ilmentym. Olio huolehtii 
*sytedatan lukemisen ja ohjaamisen suodattimien kautta muistiin ja tiedostoon. 
*olio tarjoaa mys rajapinnan muistissa olevan ja sinne talletettavan datan 
*karsimiseen toistuvasti. Sen sijaan tiedostoon talletettavan datan 
*karsintaparametrit voidaan antaa vain aloitettaessa tallennus. Edelleen 
*olio tarjoaa rajapinnan muistiin talletetun datan arvojen lukemiseen. 
*Lisksi oliolla on rajapinta dataan liittyvien otsaketietojen lukemiseen.
*/
public class Datanhallinta 
{
/*--- KONSTRUKTORIT ---*/

/*Datanhallinnasta voidaan tehd vain yksi ilmentym. Siksi konstruktori ei
*ole public-mreinen.	
*/
	protected Datanhallinta(){}
	
/*--- PUBLIC-METODIT ---*/
	
/**Datanhallinnasta voi olla vain yksi ilmentym, jonka seuraava 
*metodi palauttavaa.
*/
	public static Datanhallinta annaIlmentyma()
	{	
		if (dh==null) 
		{
			dh=new Datanhallinta();
			dh.tiedostot=new Tiedostosuodatin[MAX_TIEDOSTOJA];
		}
		return dh;
	}

/**Datanhallinnasta voi olla vain yksi ilmentym, jonka seuraava 
*metodi palauttaa.
*/
	public static Datanhallinta luoDatanhallinta()
	{ return annaIlmentyma(); }
	
/**Datanhallinnasta voi olla vain yksi ilmentym, jonka seuraava 
*metodi palauttaa.
*/	
	public static Datanhallinta annaDatanhallinta()
	{ return annaIlmentyma(); } 
	
/**Metodi on ohjelmiston tuotantoa varten.
*/
	public static Datanhallinta annaIlmentyma(boolean b)
	{	
		if (b)
		{
			if (dh==null) 
			{
				dh=new Datanhallinta();
				dh.tiedostot=new Tiedostosuodatin[MAX_TIEDOSTOJA];
			}			
		}
		else
			dh=new DatanhallintaStub();		
		return dh;
	}
	
	
/**Sulkee auki olevat tiedostot ja alustaa attribuutit.
*/
	public static void tuhoaDatanhallinta() throws IOException
	{
		dh.suljeTiedostot();
		dh.alustaAttribuutit();
		//dh=null;
	} 
	
/**Luo olion datalhde ja lukee otsikkotiedot tietorakenteeseen 
*otsikot. Tiedot luetaan tiedostosta tai putkesta. Voi aiheuttaa 
*virhetilanteita (Otsikkovirhe).
*/
	public void avaaDatalahde() throws Exception
	{
		datalahde=new Lahdetiedosto();
		alusta(datalahde);
	}
	public void avaaDatalahde(String nimi) throws Exception
	{
		//System.out.println("Datanhallinta.avaaDatalahde: Begin");	
		datalahde=new Lahdetiedosto(nimi);
		alusta(datalahde);
		//System.out.println("Datanhallinta.avaaDatalahde: End");
	}	
		
/**Sulkee datalhteen. Sulkee tiedostot.
*/
	public void suljeDatalahde() throws IOException
	{	
		datalahde.sulje();
		suljeTiedostot();
		//dh=null;	
	}
	
/**Kynnist datan lukemisen tiedostosta tai putkesta. Voi aiheuttaa
*poikkeuksia. (IOException, Datavirhe, Muistivirhe) 
*/
	public void aloitaLukeminen() throws Exception
	{
		//System.out.println("Datanhallinta.aloitaLukeminen: Begin");
		dataLuettu=false;
		boolean valmis=false;
		while (!valmis)
		try
		{
			lukusilmukka();
			valmis=true;
		}
		catch(IOException e)
		{
			System.out.println("Datanhallinta.aloitaLukeminen: IOVIRHE");	
			throw new IOvirhe("IOerror");	
		}
		catch(Exception e)
		{
			System.out.println("Datanhallinta.aloitaLukeminen: DATAVIRHE"+
			" rivill "+(rivienLkm+1));
			throw new Datavirhe("Error on dataline "+(rivienLkm+1));	
		}
		catch(OutOfMemoryError e)
		{
			System.out.println("Datanhallinta.aloitaLukeminen: MUISTIVIRHE");		
			throw new Muistivirhe("Out of memory");
		}	
		dataLuettu=true;
		//System.out.println("Datanhallinta.aloitaLukeminen: End");
	}
	
/**Palauttaa tiedon onko kaikki datalhteen data luettu.
*/
	public boolean dataLuettu()
	{ return dataLuettu; }
		
/**Asettaa tiedostosuodattimen suodatus- eli karsintatiedot. Suodattaa 
*muistissa olevan datan tiedostoon. Parametreina metodi saa tiedoston 
*nimen, tiedon kuinka paljon rivej jtetn alusta pois ja kuinka 
*tihesti rivej otetaan mukaan (suhteessa muistiin talletettuihin)
*sek niiden kenttien numerot, jotka halutaan jtt pois.
*Voi aiheuttaa poikkeuksen IOException tai IOvirhe.
*/
	synchronized public void aloitaTalletus(String nimi, int alustaPois, int riviaskel, 
		int kenttienNro[])
		throws Exception
	{
		//System.out.println("Datanhallinta.aloitatalletus: Begin");
		
		Tiedostosuodatin ts=new Tiedostosuodatin(nimi,otsikonHallinta);
		tiedostot[tiedostoja++]=ts;			
		
		ts.karsiKentat(kenttienNro);
		ts.karsiRivit(alustaPois,riviaskel);
		int absKarsittuAlusta=
			alustaPois*muisti.absKarsintaAskel()+muisti.absKarsittuAlusta();
		int absKarsintaAskel=riviaskel*muisti.absKarsintaAskel();
		
		// kirjoitetaan otsikko
		Kohdetiedosto kohde=tiedostot[tiedostoja-1].annaKohdetiedosto();
		try
		{
			otsikonHallinta.kirjoitaOtsikko(kohde,absKarsittuAlusta,
				absKarsintaAskel,kenttienNro);
		}
		catch (IOException e)
		{
			System.out.println("Datanhallinta.aloitaTalletus: IOVIRHE");
			throw e;
		}
		
		// kopioidaan muistista
		int m=muisti.annaRivienLkm();
		for(int r=0;r<m;r++)
		{
			String rivi=muisti.annaRivi(r);
			try
			{ ts.uusiRivi(rivi,r); }
			catch (Datavirhe e)
			{
				System.out.println("Datanhallinta.aloitaTalletus: DATAVIRHE");
				throw e;
			}
		}
         
		ts.karsiRivit(absKarsittuAlusta,absKarsintaAskel);
		if (dataLuettu) suljeTiedostot();
		//System.out.println("Datanhallinta.aloitatalletus: end");
	}
	synchronized public void aloitaTalletus(String nimi)
		throws Exception
	{
		int[] apu=new int[otsikonHallinta.annaKenttienLkm()];
		int j=0;
		for(int k=0;k<otsikonHallinta.annaKenttienLkm();k++)
			if (!muisti.onMuistissa(k))
				apu[j++]=k;
		int[] kenttienNro=new int[j];
		System.arraycopy(apu,0,kenttienNro,0,j);
		aloitaTalletus(nimi,0,1,kenttienNro);
	}

/**Metodi palauttaa kenttien lukumrn.*/
	public int annaKenttienLkm()
	{ return otsikonHallinta.annaKenttienLkm(); }

/**Metodi palauttaa parametrina annetun kentn nimen.*/
	public String annaKentanNimi(int kentanNro)
	{ return otsikonHallinta.annaKentanNimi(kentanNro); }

/**Metodi palauttaa true joss paarmetrina annettu kentt on funktio.*/	
	public boolean onFunktio(int kentanNro)
	{ return otsikonHallinta.onFunktio(kentanNro); }

/**Metodi palauttaa parametrina annetun kentn x-akselin nimen, jos
*kentt on funktio ja nimi on annettu lhdetiedoston otsakeosassa.
*Muuten metodi palauttaa null-arvon.
*/		
	public String annaXnimi(int kentanNro)
	{ return otsikonHallinta.annaXnimi(kentanNro); }

/**Metodi palauttaa parametrina annetun kentn mrittelyvlin
*alarajan, jos kentt on funktio ja mrittelyvli on annettu 
*lhdetiedoston otsakeosassa. Muuten metodi palauttaa NaN-arvon.
*/	
	public float annaAlaraja(int kentanNro)
	{ return otsikonHallinta.annaAlaraja(kentanNro); }

/**Metodi palauttaa parametrina annetun kentn mrittelyvlin
*ylrajan, jos kentt on funktio ja mrittelyvli on annettu 
*lhdetiedoston otsakeosassa. Muuten metodi palauttaa NaN-arvon.
*/	
	public float annaYlaraja(int kentanNro)
	{ return otsikonHallinta.annaYlaraja(kentanNro); }
	
/**Karsii parametrina annettavat kentt muistista.
*/
	public void karsiKentat(int[] kenttienNro)
	{ 
		//System.out.println("Datanhallinta.karsiKentat: begin");
		muisti.karsiKentat(kenttienNro); 
		//System.out.println("Datanhallinta.karsiKentat: end");
	}

/**Karsii muistiin talletettuja rivej. Saa parametrina tiedon kuinka 
*monta rivi alusta jtetn huomiotta ja kuinka tihesti rivej otetaan 
*sytedatasta talteen. Parametrit annetaan suhteellisina.
*/
	public int karsiRivit(int alustaPoisSuht, int riviaskelSuht)
	{
		//System.out.println("Datanhallinta.karsiRivit:");
		return muisti.karsiRivit(alustaPoisSuht, riviaskelSuht);
	}
	
/*Seuraavat metodit palauttavat tietoja muistiin talletetusta datasta 
*sek karsintasuureiden arvoja.
*/
/**Metodi palauttaa arvon, joka kertoo montako rivi muistiin on talletettu
*edellisen kutsun jlkeen.
*/
	public int uusiaRiveja() 
	{ 
		if (muisti==null) return 0; 
		return muisti.uusiaRiveja();
	}

/**Synkronoi rivien lukemista. Kutsutaan karsiRivit-kutsun yhteydess.
*/
	public void vapautaOdotus()
	{ muisti.vapautaOdotus();}
	
/**Synkronoi rivien lukemista. Kutsutaan karsiRivit-kutsun yhteydess.
*/
	public void normalisoi()
	{ muisti.normalisoi();}

/**Metodi palauttaa parametrina annetun rivin ja kentn mrmn
*data-alkion arvon.
*/
	public float annaArvo(int rivinNro, int kentanNro) 
	{ return muisti.annaArvo(rivinNro,kentanNro,1);}

/**Metodi palauttaa parametrina annetun rivin ja kentn mrmn
*data-alkion arvon. Jos kentt on funktio, tarkentimella osoitetaan 
*haluttu arvo seuraavasti: 
*0->palojen lukumr, 1->ensimminen pala, 2->ensimminen muutoskohta, 
*3->toinen pala, jne. 
*/	
	public float annaArvo(int rivinNro, int kentanNro, int tarkennin) 
	{ return muisti.annaArvo(rivinNro,kentanNro,tarkennin); }

/**Metodi palauttaa parametrina annetun rivin ja kentn mrmn
*funktion palojen lukumrn.
*/	
	public int annaPalojenLkm(int rivinNro, int kentanNro) 
	{ return muisti.annaPalojenLkm(rivinNro,kentanNro);}

/**Metodi palauttaa arvon, joka kertoo kuinka monta rivi kaiken 
*kaikkiaan on alusta ohitettu. (muistiin talletetuista riveist)	
*/	
	public int absKarsittuAlusta() 
	{  return muisti.absKarsittuAlusta() ;}

/**Metodi palauttaa arvon, joka kertoo kuinka monennent rivit kaiken 
*kaikkiaan on alkuperisest datasta talletettu muistiin.
*/	
	public int absKarsintaAskel() 
	{ return muisti.absKarsintaAskel();}
	
/**Metodi palauttaa true joss parametrina annettu kentta on talletettu 
*muistiin. Siis sit ei ole karsittu.
*/
	public boolean onMuistissa(int kentanNro) 
	{ return muisti.onMuistissa(kentanNro);}
	
/**Metodi palauttaa muistiin talletettujen rivien mrn.
*/
	public int rivejaMuistissa() 
	{ return muisti.annaRivienLkm(); }
	
/**Palauttaa datalhteest luettujen rivien mrn.
*/
	public int rivejaLuettu() 
	{ return rivienLkm; }
	
	
/*--- PROTECTED-METODIT ---*/
		
/*Lukee rivej datalhteest ja sytt ne muistisuodattimelle ja
*tiedostosuodattimelle.
*/
	protected void lukusilmukka() throws Exception 
	{
		//System.out.println("Datanhallinta.lukusilmukka: Begin");
		dataLuettu=false;
		
		char tietueidenErotin=otsikonHallinta.annaTietueidenErotin();
		String rivi="";
		while ((rivi=datalahde.lueRivi(tietueidenErotin))!=null)
		{	
			if (rivi.length()>0)
			{
				//if (rivienLkm%100==0) 
				//	System.out.println(rivienLkm+". rivi luettu");
				
				muisti.uusiRivi(rivi,rivienLkm);	
				for (int t=0;t<tiedostoja;t++)
					tiedostot[t].uusiRivi(rivi,rivienLkm);
				
				rivienLkm++;		
			}
		}		
		//System.out.println("Datanhallinta.lukusilmukka: data luettu");
		dataLuettu=true;
		suljeTiedostot();
		//System.out.println("Datanhallinta.lukusilmukka: End");
	}

	
	
/*--- PRIVATE-METODIT ---*/
	
/*Luo Otsikonhallinta-olion ja kutsuu tt lukemaan otsikkotiedot. 
*Parametrina annetaan Lahdetiedosto-luokan olio. 
*/
	private void alusta(Lahdetiedosto datalahde) throws Exception
	{
		rivienLkm=0;
		tiedostoja=0;
		otsikonHallinta=new OtsikonHallinta(datalahde);
		try
		{
			otsikonHallinta.lueOtsikot();
		}
		catch (Exception e)
		{
			System.out.println("Datanhallinta.alusta: OTSIKKOVIRHE");	
			throw new Otsikkovirhe();
		}
		muisti=new Muistisuodatin(otsikonHallinta);
	}


/*
*/ 
	synchronized private void suljeTiedostot() throws IOException
	{
		for (int t=0;t<tiedostoja;t++)
				tiedostot[t].sulje();
	
	}

/*
*/	
	synchronized private void alustaAttribuutit()
	{
		otsikonHallinta=null;
		rivienLkm=0;
		datalahde=null;
		muisti=null;
		tiedostoja=0;
		dataLuettu=false;
	}		
		
/*--- PRIVATE-TIETORAKENTEET ---*/
	
	private final static int MAX_TIEDOSTOJA=256;
		
	private OtsikonHallinta otsikonHallinta=null;
	private int rivienLkm=0;
	private Lahdetiedosto datalahde=null;
	private Muistisuodatin muisti=null;
	private Tiedostosuodatin[] tiedostot={};
	private int tiedostoja=0;
	private boolean dataLuettu=false;
	private static Datanhallinta dh=null;

	
/*--- MAIN-METODI ---*/
/*
*/
	public static void main(String[] args)
	{
		System.out.println("Datanhallinta.main: Begin");
	
		System.out.println("Datanhallinta.main: End");
	}
}

