// Tss tiedostossa on luokan Muistisuodatin toteutus.

package datanhallinta;


/**Muistisuodatin perii luokan Suodatin. Muistisuodatin tarjoaa palvelut
*datarivin lismiseen muistiin sek metodit talletettujen arvojen 
*lukemiseen. Luokkaa kytetn vain pakkauksen sislt.
*/
class Muistisuodatin extends Suodatin
{
/*--- KONSTRUKTORIT ---*/
	
/** Alustaa tietorakenteet.
*/
	public Muistisuodatin(OtsikonHallinta ots)
	{		
		super(ots);
		int kenttienLkm=ots.annaKenttienLkm();
		luoMuistiSuodatin(kenttienLkm);
	}
	
		
/*--- PUBLIC-METODIT ---*/
	
/**Palauttaa parametrien mrmn data-alkion arvon.
*/
	public float annaArvo(int rivinNro, int kentanNro, int tarkennin)
	{ return sailio[kentanNro].annaArvo(rivinNro,tarkennin);}
	
/**Palauttaa parametrin mrmn rivin merkkijonona sovitun formaatin
*mukaisesti.
*/
	public String annaRivi(int rivinNro)
	{ 
		StringBuffer sb=new StringBuffer();

		for(int k=0;k<otsikonHallinta.annaKenttienLkm();k++)
			if (onMuistissa(k))
				sb.append(sailio[k].annaArvoMerkkijonona(rivinNro)).append(kenttienErotin);
			else
				sb.append("000").append(kenttienErotin);

		return sb.toString();
	}
          
/**Palauttaa parametrien mrmn funktion palojen lukumrn. 
*/
	public int annaPalojenLkm(int rivinNro, int kentanNro)
	{ return sailio[kentanNro].annaPalojenLkm(rivinNro);} 
	
/**Metodi palauttaa true joss parametrina annettu kentta on talletettu 
*muistiin. Siis sit ei ole karsittu.
*/
	public boolean onMuistissa(int kentanNro)
	{ return !kenttaKarsittu[kentanNro]; }
	
	
	
/*--- PROTECTED-METODIT ---*/
	
/*Lis kentt tietorakenteeseen.
*/
	protected void lisaaKentat(String[] kentat) throws Exception
	{	
		for (int i=0;i<kentat.length;i++)
			if (!kenttaKarsittu[i])
					lisaaArvo(kentat[i],i);	
		
		rivienLkm++;
	}
	
/*Poistaa parametrina annetut kentt tietorakenteesta 
*/
	protected void poistaKentat(int[] kenttienNro)
	{
		for (int i=0;i<kenttienNro.length;i++)
			sailio[kenttienNro[i]]=null;			
		System.gc();
	}
	
/*Poistaa ensimmisen parametrin arvon verran rivej ja tiivist 
*rivej toisen parametrin arvon verran.
*/
	protected void tiivistaRivit(int alustaPoisSuht, int riviaskelSuht)
	{	
		int a=riviaskelSuht;
		int uusir=0;
		for (int k=0;k<kenttaKarsittu.length;k++)
			if (!kenttaKarsittu[k])
			{
				uusir=0;
				for (int r=alustaPoisSuht;r<rivienLkm;r+=a)
				{
					int i0=uusir/M,j0=uusir%M;
					int i=r/M,j=r%M;
					sailio[k].kopioi(r,uusir);				
					uusir++;
				}
				sailio[k].vapautaTila(uusir);	
			}
			System.gc();
		rivienLkm=uusir;	
	}
	
/*--- PRIVATE-METODIT ---*/
	
/*Metodi tekee tarvittavat alustuksen tietorakenteille. Saa parametrina
*kenttien lukumrn.
*/
	private void luoMuistiSuodatin(int kenttienLkm)
	{
		if (kenttienLkm!=kenttaKarsittu.length)
		{ System.out.println("VIRHE");}
		
		sailio=new Sarake[kenttienLkm];
		
		for (int i=0;i<kenttienLkm;i++)	
		{
			if (!kenttaKarsittu[i])
			{
				if (otsikonHallinta.onFunktio(i))
					sailio[i]=new LukulistaSarake(N);
				else
					sailio[i]=new LukuSarake(N);
			}
		}	
	}
	
	
/*Metodi lis ensimmisen parametrin ilmoittaman arvon 
*tietorakenteeseen. Toinen parametri kertoo mink  kentn arvo 
*on kyseess. 
*/
	private void lisaaArvo(String mjono, int kentanNro) throws Exception
	{ sailio[kentanNro].lisaaArvo(mjono,rivienLkm);	}
	
	
	
/*--- PRIVATE-TIETORAKENTEET ---*/
	
/*Dataalkiot talletetaan tietorakenteeseen sailio, joka on taulukko,
*jonka alkiot ovat kentn tyypist riippuen joko LukuSarake tai 
*LukulistaSarake -luokan olioita. Nill luokilla on yliluokka Sarake.
*/

	private int M=1024,N=1024; // => sailoon mahtuu 1024^2 rivi
	
	abstract class Sarake
	{
		public int annaPalojenLkm(int rivinNro){ return 0;}
		
		abstract public void lisaaArvo(String mjono,int rivienLkm) 
			throws Exception;	
		abstract public void vapautaTila(int rivinNro);
		abstract public void kopioi(int r, int uusir);
		abstract public String annaArvoMerkkijonona(int rivinNro);
		abstract public float annaArvo(int rivinNro,int tarkennin);
	
	}
	class LukuSarake extends Sarake
	{
		public LukuSarake(int n)
		{ arvo=new float[n][]; }
		
		public void lisaaArvo(String mjono,int rivienLkm)
		{	
			int j=rivienLkm/M,k=rivienLkm%M;
			if (arvo[j]==null) arvo[j]= new float[N];	
	
			arvo[j][k]=Float.valueOf(mjono).floatValue();	
				
		}	
		public void vapautaTila(int rivinNro)
		{
			int i0=rivinNro/M,i1=(rivienLkm-1)/M;
			for(int i=i0+1;i<=i1;i++) arvo[i]=null;
		}
		
		public void kopioi(int r, int uusir )
		{ 
			int i=r/M,j=r%M;
			int i0=uusir/M,j0=uusir%M;
			arvo[i0][j0]=arvo[i][j];
		}
		
		public String annaArvoMerkkijonona(int rivinNro)
		{
			int i=rivinNro/M,j=rivinNro%M;
			return Float.toString(arvo[i][j]);
		}
		
		public float annaArvo(int rivinNro,int tarkennin)
		{
			int i=rivinNro/M,j=rivinNro%M;
			try 
			{
				if (rivinNro<rivienLkm)
				
					return arvo[i][j];
				else
					System.out.println("Muistisuodatin.Lukusarake.annaArvo: "+
					" rivinNro="+rivinNro+" tarkennin="+tarkennin+
					" rivienLkm="+rivienLkm);
			
			} 
			catch (Exception e)
			{ 
				System.out.println("Muistisuodatin.Lukusarake.annaArvo: "+
					" rivinNro="+rivinNro+" tarkennin="+tarkennin+
					" rivienLkm="+rivienLkm);
			}
			System.exit(1);
			return 1;
		}
		
		protected float[][] arvo;
	
	}
	class LukulistaSarake extends Sarake	
	{
		public LukulistaSarake(int n)
		{ arvo=new float[n][][];}
		
		public void lisaaArvo(String mjono,int rivienLkm) throws Exception	
		{
			
			int lkm=0;
			int n=mjono.length();
			float[] f=new float[1024];
			int tila=0;
			StringBuffer s=new StringBuffer();
			if ((n<3)||(mjono.charAt(0)!='(')) 
				throw new Datavirhe();
			for (int i=1;i<n-1;i++)
			{
				char c=mjono.charAt(i);
				switch (tila)
				{
				case 0:// edellinen merkki erotin
					if (c!=' '){ s.append(c); tila=1; }
					break;
				
				case 1:// edellinen merkki muu
					if (c!=' '){ s.append(c); }
					else 
					{
						f[lkm++]=Float.valueOf(s.toString()).floatValue(); 
						s.setLength(0);
						tila=0; 
					}
					break;
				}	
			} 
		
			if (s.length()!=0) f[lkm++]=Float.valueOf(s.toString()).floatValue();
		
			int jj=rivienLkm/M,kk=rivienLkm%M;
			if (arvo[jj]==null) arvo[jj]= new float[N][];
		
			arvo[jj][kk]=new float[lkm];
			for (int l=0;l<lkm;l++) arvo[jj][kk][l]=f[l];
		}
		
		public int annaPalojenLkm(int rivinNro)
		{
			int i=rivinNro/M,j=rivinNro%M;		
			return (arvo[i][j].length+1)/2;
		}
		public void vapautaTila(int rivinNro)
		{
			int i0=rivinNro/M,i1=(rivienLkm-1)/M;
			for(int i=i0+1;i<=i1;i++) arvo[i]=null;
		}
		public void kopioi(int r, int uusir )
		{ 
			int i=r/M,j=r%M;
			int i0=uusir/M,j0=uusir%M;
			arvo[i0][j0]=arvo[i][j];
		}
		public String annaArvoMerkkijonona(int rivinNro)
		{
			int i=rivinNro/M,j=rivinNro%M;		
			float[] k=arvo[i][j];
			StringBuffer s=new StringBuffer("(");
			for (int ind=0;ind<k.length;ind++)
			{
				String t=Float.toString(k[ind]);
				s.append(t+" ");
			}	
			return s.append(")").toString();
		}
		
		public float annaArvo(int rivinNro,int tarkennin)
		{
			int i=rivinNro/M,j=rivinNro%M;
			try
			{
				if (rivinNro<rivienLkm)
				{
				if (tarkennin>0)
					return arvo[i][j][tarkennin-1];
				else
					return (arvo[i][j].length+1)/2;
				}
				else
					System.out.println(
						"Muistisuodatin.Lukulistasarake.annaArvo: "+
						" rivinNro="+rivinNro+" tarkennin="+tarkennin+
						" rivienLkm="+rivienLkm);
			}
			catch (Exception e)
			{
				System.out.println(
					"Muistisuodatin.Lukulistasarake.annaArvo: "+
					" rivinNro="+rivinNro+" tarkennin="+tarkennin+
					" rivienLkm="+rivienLkm);
			}
			System.exit(1);
			return 1;
		}
		
		protected float[][][] arvo;	
	}	
	
	private Sarake[] sailio;
	
	
/*--- MAIN-METODI ---*/
	
	public static void main(String[] args)
	{
		System.out.println("Muistisuodatin.main: Begin");

		System.out.println("Muistisuodatin.main: End");
	}
	
}// class Muistisuodatin
