
// Tss tiedostossa on luokan Suodatin toteutus.

package datanhallinta;

import java.io.*;

/**Suodatin on abstrakti luokka, joka sislt Muistisuodattimelle
*ja Tiedostosuodattimelle yhteisi piirteit. Suodatin tarjoaa metodit 
*uuden rivin lismiseen, rivien ja kenttien karsimiseen sek aksessori-
*metodeja.
*/

abstract class Suodatin
{
/*--- KONSTRUKTORIT-METODIT ---*/
	
/**Alustaa tietorakenteet.
*/
	public Suodatin(OtsikonHallinta otsLuk)
	{
		otsikonHallinta=otsLuk;
		kenttienLkm=otsLuk.annaKenttienLkm();
		tietueidenErotin=otsLuk.annaTietueidenErotin();
		kenttienErotin=otsLuk.annaKenttienErotin();
		
		kenttaKarsittu=new boolean[kenttienLkm];
		for (int i=0;i<kenttienLkm;i++)	
			kenttaKarsittu[i]=false;

	}
	
	
/*--- PUBLIC-METODIT ---*/
	
/**Palauttaa true, jos parametrin osoittama kentt on karsittu tai jos
*parametrin arvo liian suuri.
*/
	boolean onkoKarsittu(int kentanNro) 
	{ 
		if (kenttaKarsittu.length>kentanNro)
			return kenttaKarsittu[kentanNro];
		else
			return true;
	} 
	

/**Lis rivin suodattimen lpi. Parametrina annettaan listtv rivi
*merkkijonona ja datalhteest luettujen rivien mr.
*/
	synchronized public void uusiRivi(String rivi, int rivienLkm)
		throws Exception
	{
		//System.out.println("Suodatin.uusiRivi: Begin, rivi= "+rivi);
		
		int erotus=rivienLkm-absAlustaPois;
		
		if ((erotus>=0)&&(erotus%absRiviaskel==0))
		{
			String[] kentat=annaKentat(rivi,kenttienErotin);
			lisaaKentat(kentat);
			uusienRivienLkm++;
			notify();
			odottaa=false;
		}
		
		//System.out.println("Suodatin.uusiRivi: End");
	}
	
/**Karsii parametrina annettavat kentt. 
*/
	synchronized public void karsiKentat(int[] kenttienNro)
	{
		for (int i=0;i<kenttienNro.length;i++)
		{
			int j=kenttienNro[i];
			if (kenttaKarsittu.length>j)
				kenttaKarsittu[j]=true;
	
		}
		poistaKentat(kenttienNro);
	}
	
/** Karsii talletettuja datarivej. Parametri sAlustaPois kertoo kuinka
*monta ensimmist rivi poistetaan ja sRiviaskel kertoo tiheyden kuinka
*monta rivi jtetn jljelle.
*/
	synchronized public int karsiRivit(int sAlustaPois, int sRiviaskel)
	{
		suhtAlustaPois=sAlustaPois;
		suhtRiviaskel=sRiviaskel;
		
		tiivistaRivit(suhtAlustaPois,suhtRiviaskel);
		
		// Tss rivien jrjestys on trke.
		absAlustaPois+=absRiviaskel*suhtAlustaPois;
		absRiviaskel*=suhtRiviaskel;		
		uusienRivienLkm=0;
		return rivienLkm;
	}
	
/**Palauttaa uusien rivien lukumrn. Uusia rivej ovat ne, jotka on
*talletettu edellisen metodikutsun jlkeen.
*/
	synchronized public int uusiaRiveja() 
	{ 
		if (vapautetaan) return 0;
		int u=uusienRivienLkm; 
		if (u==0) 
		{
			odottaa=true;
			try
			{ 
				wait(); 
				if (vapautetaan) return 0;
			}
			catch (Exception e)
			{
				System.out.println("Suodatin.uusiaRiveja: wait ei onnistu "+e);
				System.exit(1);
			}
			u=uusienRivienLkm;
		}
		uusienRivienLkm=0;			
		return u; 
	}
	
/**Synkronoi sikeit.
*/
	synchronized public void vapautaOdotus()
	{
		vapautetaan=true;
		if (odottaa) notify();
	}
	
/**Synkronoi sikeit.
*/
	synchronized public void normalisoi()
	{
		vapautetaan=false;
	}
		 	
	
/**Palauttaa muuttujan absalustaPois arvon.
*/
	public int absKarsittuAlusta() { return absAlustaPois; }
	
/**Palauttaa muuttujan absRiviaskel arvon.
*/
	public int absKarsintaAskel() { return absRiviaskel; }
	
/**Palauttaa muuttujan suhtAlustaPois arvon.
*/
	public int suhtKarsittuAlusta() { return suhtAlustaPois; }
	
/**Palauttaa muuttujan suhtRiviaskel arvon.
*/
	public int suhtKarsintaAskel() { return suhtRiviaskel; }
	
/**Palauttaa karsinnan lpisseiden rivien mrn.
*/
	public int annaRivienLkm() 
	{ return rivienLkm; }
	
/*--- PROTECTED-METODIT ---*/

/*Abstrakti metodi, joka lisaa parametrina annettavat kentat aliluokasta
*riippuvalla tavalla.
*/
	abstract protected void lisaaKentat(String[] kentat)
		throws Exception;

/*Abstrakti metodi, joka poistaa parametrina annettavat kentat aliluokasta
*riippuvalla tavalla.
*/
	abstract protected void poistaKentat(int[] kenttienNro);

/*Abstrakti metodi, joka tiivist rivej parametrien mukaisesti aliluokasta
*riippuvalla tavalla.
*/
	abstract protected void tiivistaRivit(int suhtAlustaPois, 
	int suhtRiviaskel);

/*Synkronoitu metodi muuttaa muuttujaa uusienRivienLkm.
*/	
	synchronized protected void muutaUusienRivienLkm(int lisays)
	{ uusienRivienLkm+=lisays; }
	
/*--- PRIVATE-METODIT ---*/
	
/*Metodi paloittelee parametrina annettavan merkkijonon merkkijonolistaksi, 
*jonka jokainen alkio vastaa yht datakentt. (lukua tai lukulistaa l.
*funktioita) 
*/
	private String[] annaKentat(String mjono, char erotin)
	{
		//System.out.println("Suodatin.annaKentat: mjono='"+mjono+"'");
		String[] kentat=new String[kenttaKarsittu.length];
	
		int n=mjono.length(),m=kenttaKarsittu.length;
		
		int j=0;
		int alku=0;
		
		for(int i=0;i<n;)
		{
			char c=mjono.charAt(i++);
			if (c=='(')
			{
				alku=i-1;
				while((i<n)&&((c=mjono.charAt(i))!=')'))i++;
				kentat[j++]=mjono.substring(alku,++i);
			}
			else if (c!=erotin)
			{
				alku=i-1;
				while((i<n)&&((c=mjono.charAt(i))!=erotin))i++;
				kentat[j++]=mjono.substring(alku,i++);
			} 
		}
		return kentat;	
	}
	
	
/*--- PROTECTED-TIETORAKENTEET ---*/
	
	protected boolean[] kenttaKarsittu;
	protected int absAlustaPois=0;
	protected int absRiviaskel=1;
	protected int suhtAlustaPois=0;	
	protected int suhtRiviaskel=1;
	protected char tietueidenErotin='\n';
	protected char kenttienErotin=' ';
	protected int kenttienLkm;
	protected int rivienLkm;	
	protected OtsikonHallinta otsikonHallinta=null;
	
/*--- PRIVATE-TIETORAKENTEET ---*/

	private int uusienRivienLkm=0;
	private boolean odottaa=false;
	private boolean vapautetaan=false;
}

