package laskenta;
import datanhallinta.Datanhallinta;

/**Autokorrelaatiolaskenta periytyy abstraktista yliluokasta Laskenta. Se vastaa Auto-
*korrelaatiokaavion vaatimasta laskennasta*/

public class Autokorrelaatiolaskenta extends Laskenta
{
	public Autokorrelaatiolaskenta(boolean b)
	{
		super(b);
		kentta = -1; syvyys = 0;
	}
	
/**@see Jaljityslaskenta#valitseKentta*/
	public void valitseKentta(int k) { valitseKentta(k, 1); }
	public void valitseKentta(int k, int t) { kentta = k; tarkennin = t; muutos(); }
	
/**Asettaa suurimman etisyyden, jolta autokorrelaatio lasketaan.
*@see Autokorrelaatiolaskenta#annaSyvyys*/
	public void asetaSyvyys(int s) { syvyys = s; muutos(); }

/**Palauttaa asetaSyvyys-metodilla asetetun suurimman etisyyden.
*@see Autokorrelaatiolaskenta#asetaSyvyys*/
	public int annaSyvyys() { return syvyys; }

/**@see Jaljityslaskenta#annaMin*/
	public float annaMin() { return summain.annaPienin(); }

/**@see Jaljityslaskenta#annaMax*/
	public float annaMax() { return summain.annaSuurin(); }

/**@see Jaljityslaskenta#annaKeskiarvo*/
	public float annaKeskiarvo() { return summain.annaSumma() / rivejaKaytetty(); }

/**@see Jaljityslaskenta#annaKeskihajonta*/
	public float annaKeskihajonta() { return (float)Math.sqrt(annaVarianssi()); }

/**@see Jaljityslaskenta#annaVarianssi*/
	public float annaVarianssi()
	{
		double ka = annaKeskiarvo();
		return (float)(summain.annaNeliosumma() / rivejaKaytetty() - ka*ka);
	}
	
/**Palauttaa tarkasteltavan kentn etisyytt k vastaavan autokorrelaatiokertoimen*/
	public float annaAutokorrelaatio(int k)
	{
		int i = rengasix(k); int n = rivejaKaytetty() - k;
		if ((n < 1) || (k > syvyys)) return Float.NaN;
		double sum1 = loppusumma[i]; double sum2 = summain.annaSumma() - alkusumma[k];
		double nsum1 = loppuneliosumma[i];
		double nsum2 = summain.annaNeliosumma() - alkuneliosumma[k];
		return (float)((n*tulosumma[k] - sum1*sum2) /
			Math.sqrt((nsum1*n - sum1*sum1) * (nsum2*n - sum2*sum2)));
	}
	
	private int rengasix(int k) { int i = k + siirtyma; return i > syvyys ? i - syvyys : i; }
	
	protected boolean valmisLaskentaan() { return (syvyys > 1) && (kentta > -1); }
	
	protected void alustaUudelleenlaskenta()
	{
		summain = new Parametrisummain();
		alkusumma = new float[syvyys+1];
		alkuneliosumma = new double[syvyys+1];
		loppusumma = new float[syvyys+1];
		loppuneliosumma = new double[syvyys+1];
		loppuarvo = new float[syvyys+1];
		tulosumma = new double[syvyys+1];
		siirtyma = syvyys;
		for (int i = 1; i <= syvyys; i++) tulosumma[i] = 0;
	}
	
	protected void laskeUudelleenRivi(int dhr, int r)
	{
		float arvo = dh.annaArvo(dhr, kentta, tarkennin);
		if (r > 0)
		{
			siirtyma--;
			if (siirtyma == 0) siirtyma = syvyys;
			int i = rengasix(1);
			loppuarvo[i] = edellinen;
			loppusumma[i] = summain.annaSumma();
			loppuneliosumma[i] = summain.annaNeliosumma();
		}
		edellinen = arvo;
		summain.uusiArvo(arvo);
		if (r < syvyys)
		{
			alkusumma[r+1] = summain.annaSumma();
			alkuneliosumma[r+1] = summain.annaNeliosumma();
		}
		int j = ( r >= syvyys) ? syvyys : r;
		for (int k = 1; k <= j; k++)
			tulosumma[k] += ((double)arvo) * (double)loppuarvo[rengasix(k)];
	}
	
	protected void laskeLisaysrivi(int dhr, int r) { laskeUudelleenRivi(dhr, r); }
	
	private int kentta, tarkennin;
	private int syvyys;
	private Parametrisummain summain;

	private float [] alkusumma, loppusumma, loppuarvo;
	private double [] alkuneliosumma, loppuneliosumma, tulosumma;
	private float edellinen;
	private int siirtyma;
}


