package laskenta;
import datanhallinta.Datanhallinta;

/**Funktiolaskenta periytyy abstraktista yliluokasta Laskenta. Se vastaa paloittain vakioiden
*funktioiden vaatimasta laskennasta*/
public class Funktiolaskenta extends Laskenta
{
	public Funktiolaskenta(boolean b)
	{
		super(b);
		funktio = -1;
		jakoX = jakoY = 0;
		alakvantiili = ylakvantiili = Float.NaN;
		kvantiilit = false;
	}

/**Valitsee kaaviossa tarkasteltavan funktion f*/
	public void valitseFunktio(int f) { funktio = f; muutos(); }

/**Asettaa funktion mrittelyvlin [ala,yla], ja laskentapisteiden lkm:n p*/
	public void asetaXJako(int p, float ala, float yla)
	{ jakoX = p; alarajaX = ala; ylarajaX = yla; valiX = (yla - ala) / p; muutos(); }
	
/**Asettaa kvantiilien laskemista varten y-akselin alueen [ala,yla[, sek kuinka moneen palaan
*alue jaetaan (q)*/
	public void asetaYJako(int q, float ala, float yla)
	{ jakoY = q; alarajaY = ala; ylarajaY = yla; valiY = (yla - ala) / q; muutos(); }

/**Palauttaa mrittelyvlin alarajan*/
	public float annaMaarAlaraja() { return alarajaX; }
	
/**Palauttaa mrittelyvlin ylrajan*/
	public float annaMaarYlaraja() { return ylarajaX; }
	 
/**Metodilla asetetaan laskettavat kvantiilit 0<=ala,yla<=1. Jos ala tai yla on NaN,
*vastaavaa kvantiilia ei lasketa*/
	public void asetaKvantiilit(float ala, float yla)
	{
		alakvantiili = ala; ylakvantiili = yla;
		kvantiilit = !(Float.isNaN(ala) || Float.isNaN(yla));
	}
	
/**Palauttaa pienimmn kvantiiliarvon*/
	public float annaPieninKvantiili()
	{ return (float)(alarajaY + (pieninKvantiili - 0.5) * valiY); }

/**Palauttaa suurimman kvantiiliarvon*/
	public float annaSuurinKvantiili()
	{ return (float)(alarajaY + (suurinKvantiili - 0.5) * valiY); }

/**Palauttaa pienimmn keskiarvon*/
	public float annaPieninKa() { return pieninSumma / rivejaKaytetty(); }

/**Palauttaa suurimman keskiarvon*/
	public float annaSuurinKa() { return suurinSumma / rivejaKaytetty(); }

/**Palauttaa laskentapisteen p paikan x-akselilla*/
	public float annaX(int p) { return (float)(alarajaX + (p + 0.5f) * valiX); }

/**Palauttaa funktion keskiarvon laskentapisteess p*/
	public float annaKeskiarvo(int p) { return (float)(summa[p] / rivejaKaytetty()); }

/**Palauttaa funktion alakvantiilin arvon laskentapisteess p*/
	public float annaAlakvantiili(int p)
	{ return (float)(alarajaY + (kertymataulu[p].annaKvantiili(0) - 0.5) * valiY); }

/**Palauttaa funktion ylakvantiilin arvon laskentapisteess p*/
	public float annaYlakvantiili(int p)
	{ return (float)(alarajaY + (kertymataulu[p].annaKvantiili(1) - 0.5) * valiY); }

/**Palauttaa laskentapisteiden lukumrn*/	
	public int annaLkm() { return jakoX; }

/**Palauttaa palan k ( k = 0..p-1, p palojen lkm ) keskiarvon*/
	public float annaPalanKa(int k) { return (float)(palasumma[k] / rivejaKaytetty()); }

/**Palauttaa muutoskohdan m ( m = 0..p-2, p palojen lkm ) keskiarvon*/
	public float annaMuutoskohdanKa(int m) 
	{ return muutossumma[m] / rivejaKaytetty(); }
 
/**Laskee parametrien 0<=ala,yla<=1 mrmien kvantiilien tarkat arvot. Jos ala tai yla on NaN,
*vastaavan kvantiilin arvoksi tulee NaN*/
	public void laskeKvantiilit(float ala, float yla)
	{
		int r = rivejaKaytetty();
		alaQArvo = new float[jakoX];
		ylaQArvo = new float[jakoX];
		int alapos = Float.isNaN(ala) ? 0 : (int)Math.floor(ala * (r - 1));
		int ylapos = Float.isNaN(ala) ? r - 1 : (int)Math.ceil(yla * (r - 1));
		if ((r == 0) || (alapos >= ylapos))
		{
			for (int i = 0; i < jakoX; i++) alaQArvo[i] = ylaQArvo[i] = Float.NaN;
			return;
		}

		int[] minkekoix = new int[r];
		int[] maxkekoix = new int[r];
		int[] dhRivi = new int[r];
		short[] pala = new short[r];
		float[] arvo = new float[r];
		
		final int max1ix = 0, max2ix = alapos, min1ix = 0, min2ix = r - ylapos - 1;
		MaxUlkokeko maxkeko1 = new MaxUlkokeko(alapos, maxkekoix, arvo, max1ix);
		MaxUlkokeko maxkeko2 = new MaxUlkokeko(ylapos - alapos + 1,
				maxkekoix, arvo, max2ix);
		MinUlkokeko minkeko1 = new MinUlkokeko(r - ylapos - 1, minkekoix, arvo, min1ix);
		MinUlkokeko minkeko2 = new MinUlkokeko(ylapos - alapos + 1,
				minkekoix, arvo, min2ix);
		Muutostaulu muutostaulu = new Muutostaulu(jakoX, r);

// kekojen rakentaminen funktioiden ensimmisist paloista
		int dhr = -1;
		for (int i = 0; i < r; i++)
		{
			dhRivi[i] = dhr = seuraavaRivi(dhr); pala[i] = 1; 
			float y = arvo[i] = dh.annaArvo(dhr, funktio, 1);
			if (dh.annaPalojenLkm(dhr, funktio) > 1)
			{
				int p = annaXAlue(dh.annaArvo(dhr, funktio, 2));
				if (p < 0) muutostaulu.lisaa(r, 0);
				else if (p < jakoX) muutostaulu.lisaa(i, p);
			}
	
			int ylivuoto, j = i;
			boolean kaytetty = false;
			if (!maxkeko1.taysi()) { maxkeko1.lisaa(j); minkekoix[j] = r; kaytetty = true; }
			else if ((alapos > 0) && (y < arvo[ylivuoto = maxkeko1.annaHuippu()]))
			{
				maxkeko1.korvaa(max1ix, j); maxkeko1.paivitaPaikka(j);
				minkekoix[j] = r;
				y = arvo[ylivuoto]; j = ylivuoto;
			}
			if (!kaytetty) 
				if(!maxkeko2.taysi())
				{ maxkeko2.lisaa(j); minkeko2.lisaa(j); kaytetty = true; }
				else if (y < arvo[ylivuoto = maxkeko2.annaHuippu()])
				{
					maxkeko2.korvaa(max2ix, j); maxkeko2.paivitaPaikka(j);
					minkeko2.korvaa(minkekoix[ylivuoto], j); minkeko2.paivitaPaikka(j);
					j = ylivuoto;
				}
			if (!kaytetty) { minkeko1.lisaa(j); maxkekoix[j] = r; }
		}
		float aQ = Float.isNaN(ala) ? Float.NaN : arvo[minkeko2.annaHuippu()];
		float yQ = Float.isNaN(yla) ? Float.NaN : arvo[maxkeko2.annaHuippu()];


// laskentapisteiden lpikynti kekoja pivitten
		int lp = 0;
		while (muutostaulu.seuraava())
		{
			int p = muutostaulu.annaLaskentapiste();
			for (; lp < p; lp++)
			{
				alaQArvo[lp] = aQ;
				ylaQArvo[lp] = yQ;
			}
			int i = muutostaulu.annaRivi();
			int k = ++pala[i];
			float y = dh.annaArvo(dhRivi[i], funktio, k + k - 1);
			
			int q = -1;
			if (k < dh.annaPalojenLkm(dhRivi[i], funktio))
				if ((q = annaXAlue(dh.annaArvo(dhRivi[i], funktio, k + k))) < p)
				{
					System.out.println("laskeKvantiilit : muutoskohdat vrss jrjestyksess");
					System.exit(1);
				}
				else if (q >= jakoX) q = -1;
			muutostaulu.paivitaMuutoskohta(q);

			int j;
			if (y < arvo[i])
			{
				arvo[i] = y;
				if (maxkekoix[i] > ylapos)		// positio ylkvantiilin ylpuolella
				{
					minkeko1.paivitaPaikka(i);
					if (y < arvo[j = maxkeko2.annaHuippu()])
					{
						minkeko2.korvaa(minkekoix[j], i); minkeko1.korvaa(min1ix, j); 
						maxkeko2.korvaa(max2ix, i); maxkekoix[j] = r;
					}
				}
				if ((maxkekoix[i] >= alapos) && (maxkekoix[i] <= ylapos))
				{
					maxkeko2.paivitaPaikka(i);
					minkeko2.paivitaPaikka(i);
					if ((alapos > 0) && (y < arvo[j = maxkeko1.annaHuippu()]))
					{
						minkeko2.korvaa(min2ix, j); minkekoix[i] = r;
						maxkeko2.korvaa(maxkekoix[i], j); maxkeko1.korvaa(max1ix, i);
					}
				}
				if ((alapos > 0) && (maxkekoix[i] < alapos)) maxkeko1.paivitaPaikka(i);
			}
			else
			{
				arvo[i] = y;
				if (maxkekoix[i] < alapos)
				{
					maxkeko1.paivitaPaikka(i);
					if (y > arvo[j = minkeko2.annaHuippu()])
					{
						minkeko2.korvaa(min2ix, i); minkekoix[j] = r;
						maxkeko2.korvaa(maxkekoix[j], i); maxkeko1.korvaa(max1ix, j);
					}
				}
				if ((maxkekoix[i] >= alapos) && (maxkekoix[i] <= ylapos))
				{
					maxkeko2.paivitaPaikka(i);
					minkeko2.paivitaPaikka(i);
					if ((ylapos < r - 1) && (y > arvo[j = minkeko1.annaHuippu()]))
					{
						minkeko2.korvaa(minkekoix[i], j); minkeko1.korvaa(min1ix, i);
						maxkeko2.korvaa(max2ix, j); maxkekoix[i] = r; 
					}
				}
				if ((ylapos < r - 1) && (maxkekoix[i] > ylapos)) minkeko1.paivitaPaikka(i);
			}
			aQ = Float.isNaN(ala) ? Float.NaN : arvo[minkeko2.annaHuippu()];
			yQ = Float.isNaN(yla) ? Float.NaN : arvo[maxkeko2.annaHuippu()];
		}	
		for (; lp < jakoX; lp++) { alaQArvo[lp] = aQ; ylaQArvo[lp] = yQ; }
	}

/**Palauttaa alakvantiilin tarkan arvon*/
	public float annaTarkkaAlakvantiili(int p) {return alaQArvo[p];}

/**Palauttaa ylkvantiilin tarkan arvon*/
	public float annaTarkkaYlakvantiili(int p) {return ylaQArvo[p];}

	
	protected boolean valmisLaskentaan()
	{
		return (funktio > -1) && (jakoX > 0) && 
		(!kvantiilit || (jakoY > 0));
	}

	protected void alustaUudelleenlaskenta()
	{
		if (kvantiilit)
		{
			kertymataulu = new JoustavaKertymataulu[jakoX];
			for (int i = 0; i < jakoX; i++)
				kertymataulu[i] = new JoustavaKertymataulu(jakoY+2,2);
		}
		else kertymataulu = null;
		summa = new float[jakoX];
		if (kiinnitetty = (annaKarsintafunktio() == funktio))
		{
			palasumma = new float[annaPalaLkm()];
			muutossumma = new float[annaPalaLkm() - 1];
		}
		pieninSumma = suurinSumma = Float.NaN;
		pieninKvantiili = jakoY + 2; suurinKvantiili = -1;
		alitus = ylitys = false;
	}

	protected void laskeUudelleenRivi(int dhr, int r)
	{
		int k = dh.annaPalojenLkm(dhr, funktio);
		float y = dh.annaArvo(dhr, funktio, 1);
		summa[0] += y;
		if (kiinnitetty) palasumma[0] += y;
		int q = annaYAlue(y);
		if (kvantiilit) kertymataulu[0].kasvataArvoa(q);
		float x0 = -Float.MAX_VALUE;
		for (int i = 0; i < k - 1; i++)
		{
			float x = dh.annaArvo(dhr, funktio, i + i + 2);
			int p = annaXAlue(x);
			if (x < x0)
			{
				System.out.println("Funktiolaskenta.laskeUudelleenrivi: datavirhe, " +
				"muutoskohdat vrss jrjestyksess, rivi = " + dhr);
				System.exit(1);
			}
			x0 = x;
			if (p < 0) { p = 0; alitus = true; }
			else if (p >= jakoX) ylitys = true;  
 			float uusiY = dh.annaArvo(dhr, funktio, i + i + 3);
			if (p < jakoX)
			{
				if (kvantiilit) kertymataulu[p].pienennaArvoa(q);
				summa[p] += (uusiY - y);
				q = annaYAlue(uusiY);
				if (kvantiilit) kertymataulu[p].kasvataArvoa(q);
			}
			y = uusiY;
			if (kiinnitetty) { palasumma[i+1] += y; muutossumma[i] += x; }
		}
	}

	protected void viimeisteleUudelleenlaskenta(int r)
	{
		for (int p = 1; p < jakoX; p++) summa[p] += summa[p-1];
		if (kvantiilit)
		{
			for (int p = 1; p < jakoX; p++)
				for (int q = 0; q < jakoY + 2; q++)
					kertymataulu[p].asetaArvo(q, kertymataulu[p].annaArvo(q) +
					kertymataulu[p-1].annaArvo(q));
			for (int p = 0; p < jakoX; p++) kertymataulu[p].valmistele();
		}
		viimeisteleLisayslaskenta(r);
	}

	protected void laskeLisaysrivi(int dhr,int r)
	{
		float x,y = 0;
		float x0 = -Float.MAX_VALUE;
		int k = dh.annaPalojenLkm(dhr, funktio);
		if (kiinnitetty && (k != annaPalaLkm()))
		{ System.out.println("vr palalkm="+k+",dhr="+dhr); System.exit(1); }
		int p = 0; int m = 0;
		for (int i = 0; i < jakoX; i++)
		{
			while (i == p)
			{
				y = dh.annaArvo(dhr,funktio,m+m+1);
				if (m < k - 1)
				{
					x = dh.annaArvo(dhr,funktio,m+m+2);
					if (x < x0)
					{ /* error */
						System.out.println("Funktiolaskenta.laskeLisaysrivi: "+
						"datavirhe, muutoskohdat vrss jrjestyksess, rivi = " + dhr);
						System.exit(1);
					}
					x0 = x;
					p = annaXAlue(x);
					if (p < 0) { p = 0; alitus = true; }
					else if (p >= jakoX) ylitys = true;
					if (kiinnitetty) muutossumma[m] += x;
				}
				else p = jakoX;
				if (kiinnitetty) palasumma[m] += y;
				m++;
			}
			summa[i] += y;
			if (kvantiilit) kertymataulu[i].uusiArvo(annaYAlue(y));
		}
	}

	protected void viimeisteleLisayslaskenta(int r)
	{
		for (int p = 0; p < jakoX; p++)
		{
			if (kvantiilit)
			{
				kertymataulu[p].paivitaKvantiili(0,(int)Math.floor(alakvantiili*(r - 1)));
				kertymataulu[p].paivitaKvantiili(1,(int)Math.ceil(ylakvantiili*(r - 1)));
				int k0 = kertymataulu[p].annaKvantiili(0);
				int k1 = kertymataulu[p].annaKvantiili(1);
				if (k0 > k1) { int k = k0; k0 = k1; k1 = k; }
				if (Float.isNaN(pieninSumma)) { pieninKvantiili = k0; suurinKvantiili = k1; }
				else if (k0 < pieninKvantiili) pieninKvantiili = k0;
				else if (k1 > suurinKvantiili) suurinKvantiili = k1;
			}
			float s = summa[p];
			if (Float.isNaN(pieninSumma)) pieninSumma = suurinSumma = s;
			else
			{
				if (s < pieninSumma) pieninSumma = s;
				else if (s > suurinSumma) suurinSumma = s;
			}
		}
	}
	
	private int annaYAlue(float y)
	{
		if (y < alarajaY) return 0;
		if (y >= ylarajaY) return jakoY + 1;
		return 1 + (int)((y - alarajaY) / valiY);
	}
	
	private int annaXAlue(float x) { return (int)Math.floor((x - alarajaX) / valiX); }

	private JoustavaKertymataulu[] kertymataulu;
	private float[] summa, palasumma, muutossumma, alaQArvo, ylaQArvo;
	private int funktio;
	private int jakoX, jakoY;
	private boolean kiinnitetty, kvantiilit;
	private boolean alitus, ylitys;
	private float alakvantiili, ylakvantiili;
	private float alarajaX, alarajaY, ylarajaX, ylarajaY, valiX, valiY;
	private float pieninSumma, suurinSumma;
	private int pieninKvantiili, suurinKvantiili;
}
