package laskenta;
import datanhallinta.Datanhallinta;

/**Jakaumalaskenta periytyy abstraktista Laskenta-yliluokasta. Se vastaa parametrin
*jakaumaa kuvaavan histogrammikaavion vaatimasta laskennasta. */
public class Jakaumalaskenta extends Laskenta
{
	public Jakaumalaskenta(boolean b)
	{
		super(b);
		kentta = -1; pylvaita = 0; tarkatKvantiilit = false;
		alakvantiili = ylakvantiili = Float.NaN;
	}
	
/**@see Jaljityslaskenta#valitseKentta*/
	public void valitseKentta(int k) { valitseKentta(k,1); }
	public void valitseKentta(int k,int t) { kentta = k; tarkennin = t; muutos(); }

/**Asettaa histogrammipylviden lukumrn lkm, ja arvoalueen [ala,yla[ */
 	public void asetaPylvaat(int lkm,float ala,float yla)
	{
		pylvaita = lkm; leveys = (yla-ala) / lkm;
		alaraja = ala; ylaraja = yla;
		muutos();
	}

/**Palauttaa pylviden lukumrn*/
	public int annaLkm() { return pylvaita; }

/**Metodilla asetetaan laskettavat kvantiilit 0<=ala,yla<=1. Jos ala tai yla on NaN,
*kvantiilia ei lasketa*/
	public void asetaKvantiilit(float ala,float yla) { alakvantiili = ala; ylakvantiili = yla; }

/** */
	public void laskeTarkatKvantiilit() { tarkatKvantiilit = true; }
		
/**Palauttaa pylvn p alueelle osuneiden parametrin arvojen kertymn, jos 1<=p<=P,
*jossa P on pylviden lkm. Jos p = 0, palauttaa pylviden kattaman arvoalueen
*alapuolisen kertymn, ja jos p = P+1, ylpuolisen kertymn. */
	public int annaKertyma(int p) { return kertymataulu.annaKertyma(p); }

/**Palauttaa suurimman kertymn. Ei ota huomioon arvoalueen ulkopuolisia kertymi */
	public int annaMaxKertyma() { return suurinKertyma; }

/**Palauttaa pylvn p arvoalueen ylrajan, jos p>0. Jos p=0, palauttaa pylvn 1
*arvoalueen alarajan.*/
	public float annaX(int p) { return (float)(alaraja + p*leveys); }

/**Palauttaa pylvn p = 1..P arvoalueen keskiarvon*/
	public float annaKeskiX(int p)
	{ return 0.5f*(annaX(p - 1) + annaX(p)); }
	
/**Palauttaa alakvantiilin sisltvn pylvn numeron*/
	public int annaAlakvantiili() { return kertymataulu.annaKvantiili(0); }

/**Palauttaa mediaanin sisltvn pylvn numeron*/
	public int annaMediaani() { return kertymataulu.annaKvantiili(1); }

/**Palauttaa ylkvantiilin sisltvn pylvn numeron*/
	public int annaYlakvantiili() { return kertymataulu.annaKvantiili(2); }

/**@see Jaljityslaskenta#annaMin*/
	public float annaMin() { return summain.annaPienin(); }

/**@see Jaljityslaskenta#annaMax*/
	public float annaMax() { return summain.annaSuurin(); }

/**@see Jaljityslaskenta#annaKeskiarvo*/
	public float annaKeskiarvo() { return (float)summain.annaSumma() / rivejaKaytetty(); }

/**@see Jaljityslaskenta#annaKeskihajonta*/
	public float annaKeskihajonta() {return (float)Math.sqrt(annaVarianssi());}

/**@see Jaljityslaskenta#annaVarianssi*/
	public float annaVarianssi()
	{
		double ka = annaKeskiarvo();
		return (float)(summain.annaNeliosumma() / rivejaKaytetty() - ka*ka);
	}

/**laskee parametrien ala,yla mrmien kvantiilien ja mediaanin tarkat arvot*/
	public void laskeKvantiilit(float ala, float yla)
	{
		int t, r = rivejaKaytetty();
		int[] pos = new int[3], pylvas = new int[3], jarj = {0,1,2};
		float[] kvant = new float[3];
		MaxKeko[] keko = new MaxKeko[3];
		if (r == 0) { tAlakvantiili = tMediaani = tYlakvantiili = Float.NaN; return; }

// lasketaan kvantiilipositiot (0 .. r - 1, r = rivien lkm)
		pos[0] = Float.isNaN(ala) ? 0 : (int)Math.floor((r - 1) * ala); 
		pos[1] = (int)Math.round((r - 1) * 0.5);
		pos[2] = Float.isNaN(yla) ? 0 : (int)Math.ceil((r - 1) * yla);

// kvantiilipositiot jrjestetn nousevaan jrjestykseen.
// tarpeen, koska esim. alakvantiili voi olla > 0.5
		if (pos[1] > pos[2])
		{ t = pos[1]; pos[1] = pos[2]; pos[2] = t; t = jarj[1]; jarj[1] = jarj[2]; jarj[2] = t; }
		if (pos[0] > pos[1])
		{ t = pos[0]; pos[0] = pos[1]; pos[1] = t; t = jarj[0]; jarj[0] = jarj[1]; jarj[1] = t; }
		if (pos[1] > pos[2])
		{ t = pos[1]; pos[1] = pos[2]; pos[2] = t; t = jarj[1]; jarj[1] = jarj[2]; jarj[2] = t; }

// etsitn pylvt, joiden alueelle kvantiilit kuuluvat (pylvas[j])
// lasketaan kvantiilin positio pylvn sisll (pos[j] = 1 .. p, p = pylvn kertym)
		for (int i = 0, j = 0 , k = 0; (i < pylvaita + 2) && (j < 3); i++)
		{
			int p = kertymataulu.annaKertyma(i);
			while ((j < 3) && (pos[j] >= k) && (pos[j] < k+p))
			{
				pos[j] -= k - 1; pylvas[j] = i;
				j++;
			}
			k += p;
		}
		
		
// kahden kvantiiliposition osuessa saman pylvn alueelle,
// jlkimmist vastaavan keon kooksi tulee positioiden erotus (voi olla 0)
		if (pylvas[1] == pylvas[2]) pos[2] -= pos[1];
		if (pylvas[0] == pylvas[1]) pos[1] -= pos[0];
		

// kekojen rakentaminen
		for (int j = 0; j < 3; j++) keko[j] = new MaxKeko(pos[j]);
		int dhr = -1;
		while ((dhr = seuraavaRivi(dhr)) > -1)
		{
			float x = dh.annaArvo(dhr,kentta,tarkennin);
			int p = annaPylvas(x);
			for (int j = 0; j < 3; j++)
			{
				float y;
				if (p != pylvas[j]) continue;
				if (!keko[j].taysi()) { keko[j].lisaa(x); p = -1; }
				else if ((pos[j] > 0) && (x < (y = keko[j].annaMax())))
				{
					keko[j].korvaaMax(x);
					x = y;	// keon ylivuoto mahdollisesti seuraavan kvantiilin kekoon
				}
			}
		}

// asetetaan tulosmuuttujat
		float x = 0;		// tarpeeton alustus, mutta kntj vaatii sen
		for (int j = 0; j < 3; j++)
		{
			if (pos[j] > 0) x = keko[j].annaMax();	// pos[0] > 0 aina
			kvant[jarj[j]] = x;
		}
		tAlakvantiili = Float.isNaN(ala) ? Float.NaN : kvant[0];
		tMediaani = kvant[1];
		tYlakvantiili = Float.isNaN(yla) ? Float.NaN : kvant[2];
		for (int j = 0; j < 3; j++) keko[j] = null;
		keko = null; 
	}
	
/**Palauttaa tarkan alakvantiilin arvon
*@see Jakaumalaskenta#laskeKvantiilit*/
	public float annaTarkkaAlakvantiili() { return tAlakvantiili; }

/**Palauttaa tarkan mediaanin arvon
*@see Jakaumalaskenta#laskeKvantiilit*/
	public float annaTarkkaMediaani() { return tMediaani; }

/**Palauttaa tarkan ylkvantiilin arvon
*@see Jakaumalaskenta#laskeKvantiilit*/
	public float annaTarkkaYlakvantiili() { return tYlakvantiili; }

	protected boolean valmisLaskentaan() { return (kentta > -1) && (pylvaita > 0); }
	
	protected void alustaUudelleenlaskenta()
	{
		kertymataulu = new KiinteaKertymataulu(pylvaita + 2, 3);
		summain = new Parametrisummain();
		suurinKertyma = 0;
	}

	protected void laskeUudelleenRivi(int dhr, int r)
	{
		float x = dh.annaArvo(dhr,kentta,tarkennin);
		summain.uusiArvo(x);
		int p = annaPylvas(x);
		kertymataulu.uusiArvo(p);
		int k = kertymataulu.annaKertyma(p);
		if ((k > suurinKertyma) && (p > 0) && (p <= pylvaita))
			suurinKertyma = k;
	}

	protected void laskeLisaysrivi(int dhr, int r) { laskeUudelleenRivi(dhr,r); }
	
	protected void viimeisteleUudelleenlaskenta(int r)
	{
		if (r == 0) return;
		if (!Float.isNaN(alakvantiili))
			kertymataulu.paivitaKvantiili(0,(int)Math.floor(alakvantiili*(r - 1)));
		kertymataulu.paivitaKvantiili(1,(int)(0.5*(r - 1)));
		if (!Float.isNaN(ylakvantiili))
			kertymataulu.paivitaKvantiili(2,(int)Math.ceil(ylakvantiili*(r - 1)));
		if (tarkatKvantiilit)
		{
			laskeKvantiilit(alakvantiili, ylakvantiili);
			tarkatKvantiilit = false;
		}
	}
	
	protected void viimeisteleLisayslaskenta(int r) { viimeisteleUudelleenlaskenta(r); }

// Palauttaa arvoa x vastaavan pylvn numeron
	private int annaPylvas(float x)
	{
		if (x < alaraja) return 0;
		if (x >= ylaraja) return pylvaita + 1;
		return 1 + (int)Math.floor((x - alaraja) / leveys);
	}

	private KiinteaKertymataulu kertymataulu;
	private Parametrisummain summain;
	private int kentta, tarkennin, pylvaita, suurinKertyma;
	private float alakvantiili, ylakvantiili, alaraja, ylaraja, leveys;
	private float tAlakvantiili, tYlakvantiili, tMediaani;
	private boolean tarkatKvantiilit;
}

