package laskenta;
import datanhallinta.Datanhallinta;

/**Jaljityslaskenta periytyy abstraktista Laskenta-yliluokasta. Se vastaa simulaation
*jljityksen vaatimasta laskennasta. */

public class Jaljityslaskenta extends Laskenta
{
	public Jaljityslaskenta(boolean b)
	{
		super(b);
		kentta = -1;
	}
	
/**Valitsee kaaviossa tarkasteltavan kentn k=0..N-1, jossa N on sytteen kenttien lkm.
*Tarkennin t annetaan vain, jos kentt on funktio. Se mr, mit arvoa funktiosta
*tarkastellaan:
*0: palojen lukumr (vain, jos sit ei ole kiinnitetty)
*1: ensimminen taso, 2: ensimminen muutoskohta, 3: toinen taso, jne. (vain, jos palamr
*on kiinnitetty)*/
	public void valitseKentta(int k) { valitseKentta(k, 1); }
	public void valitseKentta(int k, int t) 
	{ kentta = k; tarkennin = t; muutos(); }

/**Palauttaa kentn pienimmn arvon lasketuilta riveilt.*/
	public float annaMin() { return summain.annaPienin(); }

/**Palauttaa kentn suurimman arvon lasketuilta riveilt.*/
	public float annaMax() { return summain.annaSuurin(); }

/**Palauttaa kentn arvojen keskiarvon lasketuilta riveilt.*/
	public float annaKeskiarvo() { return summain.annaSumma() / rivejaKaytetty(); }

/**Palauttaa kentn arvojen keskihajonnan lasketuilta riveilt.*/
	public float annaKeskihajonta() { return (float)Math.sqrt(annaVarianssi()); }

/**Palauttaa kentn arvojen varianssin lasketuilta riveilt.*/
	public float annaVarianssi()
	{
		double ka = annaKeskiarvo();
		return (float)(summain.annaNeliosumma() / rivejaKaytetty() - ka*ka);
	}

/**Asettaa riviosoittimen osoittamaan suhteellista rivi r vastaavaan datanhallinnan riviin.
*@see Jaljityslaskenta#annaY
*@see Jaljityslaskenta#siirraRiviosoitinta*/
	public void asetaRiviosoitin(int r) { riviosoitin = annaDhRivi(r); }

/**Palauttaa tarkasteltavan kentn arvon riviosoittimen osoittamalta rivilt.
*@see Jaljityslaskenta#asetaRiviosoitin
*@see Jaljityslaskenta#siirraRiviosoitinta*/
	public float annaY()
	{
		if (onMuutettu())
		{
			System.out.println("Jaljityslaskenta.annaY: uudelleenlaskenta suorittamatta");
			System.exit(1);
		}
		float arvo = dh.annaArvo(riviosoitin,kentta,tarkennin);
		return arvo;
	}

/**Siirt riviosoittimen osoittamaan seuraavaan karsinnan lpisevn riviin.
*@see Jaljityslaskenta#asetaRiviosoitin
*@see Jaljityslaskenta#annaY*/
	public void siirraRiviosoitinta()
	{
		riviosoitin = seuraavaRivi(riviosoitin);
	}
				
	protected boolean valmisLaskentaan()
	{
		return kentta > -1;
	}

	protected void alustaUudelleenlaskenta() { summain = new Parametrisummain(); }
	protected void laskeUudelleenRivi(int dhr, int r)
	{
		float arvo = dh.annaArvo(dhr, kentta, tarkennin);
		summain.uusiArvo(arvo);
	}

	protected void laskeLisaysrivi(int dhr, int r) { laskeUudelleenRivi(dhr, r); }

	private int riviosoitin, kentta, tarkennin;
	Parametrisummain summain;
}


