package laskenta;

/**JoustavaKertymataulu periytyy abstraktista yliluokasta Kertymataulu. Sit kytetn 
*paloittain vakioiden funktioiden likimristen kvantiilien laskemisessa. Jokaista laskenta-
*pistett kohti varataan yksi ilmentym JoustavaKertymataulu-luokasta. Tilankytn
*optimoimiseksi taulun alkioille varataan aluksi 16-bittinen short, ja ylivuodon sattuessa 
*(alkion kertym ylitt 32767) 32-bittinen int. **/
class JoustavaKertymataulu extends Kertymataulu
{
/**Konstruktori saa parametreina taulukon koon ja laskettavien kvantiilien lukumrn*/
	protected JoustavaKertymataulu(int koko, int kvantiililkm)
	{
		super(kvantiililkm);
		taulu = new short[koko];
		for (int i = 0; i < koko; i++ ) taulu[i] = 0;
		ylityksia = 0;
		jatkepuu = null;
	}

	protected int annaArvo(int p)
	{
		short i = taulu[p];
		return (i > RAJA) ? i : annaSuuriArvo(i);
	}

	protected void asetaArvo(int p, int x)
	{
		short i = taulu[p];
		if (i <= RAJA) asetaSuuriArvo(i, x);
		else if ((x <= MAXSHORT) && (x > RAJA))
			taulu[p] = (short)x;
		else varaaUusiJatke(p, x);
	}
	
	protected void kasvataArvoa(int p)
	{
		short i = taulu[p];
		if ((i > RAJA) && (i < MAXSHORT)) taulu[p]++;
		else asetaArvo(p, annaArvo(p)+1);
	}
	
	protected void pienennaArvoa(int p)
	{
		if (taulu[p] > RAJA+1) taulu[p]--;
		else asetaArvo(p, annaArvo(p)-1);
	}

// Ylivuodon sattuessa
	private void varaaUusiJatke(int p, int x)
	{
		taulu[p] = (short)(MINSHORT + ylityksia);
		Solmu solmu = jatkepuu;
		if ((ylityksia == 64) || (ylityksia == 8))
		{
			Haara haara = new Haara();
			haara.poika[0] = jatkepuu;
			jatkepuu = solmu = haara;
		}
		if (ylityksia > 63) 
		{
			if ((ylityksia & 63) == 0) solmu = ((Haara)solmu).poika[ylityksia >> 6]
						= new Solmu();
			else solmu = ((Haara)solmu).poika[ylityksia >> 6];
		}
		if (ylityksia > 7)
		{
			if ((ylityksia & 7) == 0) solmu = ((Haara)solmu).poika[ylityksia >> 3]
						= new Lehti();
			else solmu = ((Haara)solmu).poika[ylityksia >> 3];	
		}
		if (ylityksia == 0)
		{
			Lehti lehti = new Lehti();
			jatkepuu = solmu = lehti;
		}
		((Lehti)solmu).kertyma[ylityksia & 7] = x;
		ylityksia++;
	}

	private int annaSuuriArvo(int i)
	{
		Lehti lehti = haeLehti(i-MINSHORT);
		return lehti.kertyma[i & 7];
	}

	private void asetaSuuriArvo(int i,int kertyma)
	{
		Lehti lehti = haeLehti(i-MINSHORT);
		lehti.kertyma[i & 7] = kertyma;
	}

	private Lehti haeLehti(int i)
	{
		Solmu solmu = jatkepuu;
		if (ylityksia > 63) { solmu = ((Haara)solmu).poika[i >> 6]; i &= 63; }
		if (ylityksia > 7) solmu = ((Haara)solmu).poika[i >> 3];
		return (Lehti)solmu;
	}

	private class Solmu {}
	private class Haara extends Solmu { protected Solmu[] poika = new Solmu[8]; }
	private class Lehti extends Solmu { protected int[] kertyma = new int[8]; }

	private short [] taulu;
	private Solmu jatkepuu;
	private final short RAJA = (short)0xC000, MINSHORT = (short)0x8000,
			    MAXSHORT = (short)0x7FFF;
	private short ylityksia;
}


