package laskenta;
import datanhallinta.Datanhallinta;
/**Tss abstraktissa yliluokassa on toteutettu laskennan kaikille kaaviotyypeille yhteiset *piirteet. Paikallinen karsinta on toteutettu kokonaan tss luokassa.*/

abstract class Laskenta
{
	protected Laskenta(boolean b)
	{
		karsiAlusta = 0; karsintaAskel = 1;
		karsintafunktio = -1;
		dh = Datanhallinta.annaIlmentyma(b);
		muutos();
	}

/**Jos aliluokan valmisLaskentaan-metodi palauttaa true,
*laskee kaikki rivit parametrina annettuun datanhallinnan riviin asti,
*ja palauttaa true, muuten palauttaa false.
*Kytt pakkauksen sisist, parametritonta laskeUudelleen-metodia*/
	public boolean laskeUudelleen(int dhRiveja)
	{
		viimDhRivi = dhRiveja - 1;
		return laskeUudelleen();
	}

/**Template-metodi uudelleenlaskentaan, kytt aliluokissa mriteltyj metodeita
*valmisLaskentaan, alustaUudelleenLaskenta, laskeUudelleenRivi ja
*viimeisteleUudelleenlaskenta, sek seuraavaRivi-metodia.
*@see valmisLaskentaan.
*@see alustaUudelleenLaskenta
*@see laskeUudelleenRivi
*@see viimeisteleUudelleenlaskenta
*@see seuraavaRivi
*@see laskeLisays
*/
	public boolean laskeUudelleen()
	{
		if (!valmisLaskentaan())
		{
			System.out.println("Laskenta.laskeUudelleen: ei valmis laskentaan");
			return false;
		}
		alustaUudelleenlaskenta();
		int dhr = rivi = viimLaskettuDhRivi = -1;
		while ((dhr = seuraavaRivi(dhr)) != -1)
		{
			laskeUudelleenRivi(dhr, ++rivi);
			viimLaskettuDhRivi = dhr;
		}
		viimeisteleUudelleenlaskenta(rivi + 1);
		muutettu = false;
		return true;
	}

/**Laskee parametrina annetun mrn uusia rivej.
*Template-metodi, kytt aliluokissa mriteltyj metodeita
*laskeLisysrivi ja viimeisteleLisyslaskenta, sek seuraavaRivi-metodia.
*@see laskeLisaysrivi
*@see viimeisteleLisayslaskenta
*@see seuraavaRivi
*@see laskeUudelleen
*/
	public void laskeLisays(int uusiaRiveja)
	{
		if (muutettu)
		{
			System.out.println("Laskenta.laskeLisays: muutettu = true");
			System.exit(1);
		}
		int dhr = viimLaskettuDhRivi;
		viimDhRivi += uusiaRiveja;
		alustaLisayslaskenta();
		while ((dhr = seuraavaRivi(dhr)) != -1)
		{
			laskeLisaysrivi(dhr, ++rivi);
			viimLaskettuDhRivi = dhr;
		}
		viimeisteleLisayslaskenta(rivi + 1);
	}

/**Asettaa rivien paikallisen karsinnan. Laskennasta jtetn pois alusta rivi,
*ja lopuista lasketaan joka n:s rivi, jossa n on askel. karsiRivit(0,1) poistaa karsinnan.
*@see Laskenta#kiinnitaPalaLkm
*@see Laskenta#vapautaPalaLkm
*/
	public void karsiRivit(int alusta,int askel)
	{
		karsiAlusta = alusta;
		karsintaAskel = askel;
		muutos();
	}

/**Metodi kiinnittaa parametrina annetun funktion palojen lukumrn. Laskennassa kytetn
*vain niit rivej, jotka ovat lpisseet karsiRivit-metodilla asetetun karsinnan ja joilla
*parametrina annettu funktio sislt lkm palaa.
*@see Laskenta#vapautaPalaLkm
*@see Laskenta#karsiRivit
*/
	public void kiinnitaPalaLkm(int funktio, int lkm)
	{
		karsintafunktio = funktio;
		palalkm = lkm;
		muutos();
	}

/**Metodi poistaa kiinnitaPalaLkm-metodilla asetetun karsinnan.
*@see Laskenta#kiinnitaPalaLkm
*@see Laskenta#karsiRivit
*/
	public void vapautaPalaLkm() { karsintafunktio = -1; muutos(); }

/**Palauttaa funktion, jonka palamr on kiinnitetty, tai -1.
*@see Laskenta#kiinnitaPalaLkm
*/
	public int annaKarsintafunktio() { return karsintafunktio; }

/**Palauttaa kiinnitetyn palamrn
*@see Laskenta#kiinnitaPalaLkm
*/
	public int annaPalaLkm() { return palalkm; }
	
/**Palauttaa laskennassa kytettyjen rivien lukumrn*/
	public int rivejaKaytetty() { return rivi + 1; }

/**Palauttaa parametrina annettua datanhallinnan rivi dhr seuraavan paikallisen karsinnan
*lpisevn rivin numeron. Jos dhr = -1, palauttaa ensimmisen karsinnan lpisevn rivin
*numeron. Jos annettu rivi on viimeinen karsinnan lpisev, palauttaa -1.*/
	protected int seuraavaRivi(int dhr)
	{
		dhr = (dhr == -1) ? karsiAlusta : dhr + karsintaAskel;
		while ((dhr <= viimDhRivi) && (karsintafunktio != -1) &&
		(dh.annaPalojenLkm(dhr, karsintafunktio) != palalkm)) dhr += karsintaAskel;
		return (dhr > viimDhRivi) ? -1 : dhr;
	}

/**Palauttaa suhteellista rivinumeroa rivi vastaavan datanhallinnan rivinumeron.
*Kytt seuraavaRivi-metodia, jos funktion palalukumr on kiinnitetty.
*@see seuraavaRivi
*@see kiinnitaPalaLkm
*@see vapautaPalaLkm
*/
	protected int annaDhRivi(int rivi)
	{
		int dhr = -1;
		if (karsintafunktio == -1)
		{
			dhr = karsiAlusta + rivi*karsintaAskel;
			if (dhr > viimDhRivi)
			{
				System.out.println("Laskenta.annaDhRivi: liian suuri "+
				"rivinumero : "+rivi);
				System.exit(1);
			}
		}
		else for ( ; rivi >= 0; rivi--)
		{
			dhr = seuraavaRivi(dhr);
			if (dhr == -1)
			{
				System.out.println("Laskenta.annaDhRivi: liian "+
				"suuri rivinumero");
				System.exit(1);
			}
		}
		return dhr;
	}

/** Tt metodia kutsutaan, kun laskennan parametreja muutetaan */
	protected void muutos() { muutettu = true; }

/** Palauttaa true, jos laskennan parametreja on muutettu*/
	protected boolean onMuutettu() { return muutettu; }

/**Aliluokissa toteutettava metodi, palauttaa true, jos laskentaan vaikuttavat parametrit on
*asetettu niin, ett laskenta on mahdollista. */
	abstract protected boolean valmisLaskentaan();

	abstract protected void alustaUudelleenlaskenta();
	abstract protected void laskeUudelleenRivi(int dhRivi, int rivi);
	abstract protected void laskeLisaysrivi(int dhRivi, int rivi);
		
	protected void alustaLisayslaskenta() {};
	protected void viimeisteleUudelleenlaskenta(int rivi) {};
	protected void viimeisteleLisayslaskenta(int rivi) {};
	
	
	protected Datanhallinta dh;
	private int viimDhRivi, viimLaskettuDhRivi, rivi;
	private int karsiAlusta, karsintaAskel;
	private int karsintafunktio, palalkm;
	private boolean muutettu;
}

