package laskenta;
import datanhallinta.Datanhallinta;

/**Yhteisjakaumalaskenta periytyy abstraktista yliluokasta Laskenta. se vastaa kahden
*parametrin yhteisjakaumakaavion vaatimasta laskennasta.*/

public class Yhteisjakaumalaskenta extends Laskenta
{
	public Yhteisjakaumalaskenta(boolean b)
	{
		super(b);
		kenttaX = kenttaY = -1;
	}

/**Valitsee x-akselilla tarkasteltavan kentn.
*@see Jaljityslaskenta#valitseKentta*/	
	public void valitseXKentta(int k) { valitseXKentta(k, 1); }
	public void valitseXKentta(int k, int t) { kenttaX = k; tarkenninX = t; muutos(); }
	
/**Valitsee y-akseliila tarkasteltavan kentn.
*@see Jaljityslaskenta#valitseKentta*/
	public void valitseYKentta(int k) { valitseYKentta(k, 1); }
	public void valitseYKentta(int k, int t){ kenttaY = k; tarkenninY = t; muutos(); }

/**@see Jaljityslaskenta#annaMin*/
	public float annaXMin() { return summainX.annaPienin(); }

/**@see Jaljityslaskenta#annaMax*/
	public float annaXMax() { return summainX.annaSuurin(); }

/**@see Jaljityslaskenta#annaKeskiarvo*/
	public float annaXKeskiarvo() { return summainX.annaSumma() / rivejaKaytetty(); }

/**@see Jaljityslaskenta#annaKeskihajonta*/
	public float annaXKeskihajonta() { return (float)Math.sqrt(annaXVarianssi()); }

/**@see Jaljityslaskenta#annaVarianssi*/
	public float annaXVarianssi()
	{
		double ka = annaXKeskiarvo();
		return (float)(summainX.annaNeliosumma() / rivejaKaytetty() - ka*ka);
	}

/**@see Jaljityslaskenta#annaMin*/
	public float annaYMin() { return summainY.annaPienin(); }

/**@see Jaljityslaskenta#annaMax*/
	public float annaYMax() { return summainY.annaSuurin(); }

/**@see Jaljityslaskenta#annaKeskiarvo*/
	public float annaYKeskiarvo() { return summainY.annaSumma() / rivejaKaytetty(); }

/**@see Jaljityslaskenta#annaKeskihajonta*/
	public float annaYKeskihajonta() { return (float)Math.sqrt(annaYVarianssi()); }

/**@see Jaljityslaskenta#annaVarianssi*/
	public float annaYVarianssi()
	{
		float ka = annaYKeskiarvo();
		return (float)(summainY.annaNeliosumma() /
			rivejaKaytetty() - (double)ka * (double)ka);
	}

/**Palauttaa kenttien arvojen korrelaatiokertoimen lasketuilta riveilt*/
	public float annaKorrelaatio()
	{
		int n = rivejaKaytetty();
		double sumX = summainX.annaSumma();
		double sumY = summainY.annaSumma();
		return (float)((n*tulosumma - sumX*sumY) /
			Math.sqrt((summainX.annaNeliosumma()*n - sumX*sumX)
				*(summainY.annaNeliosumma()*n - sumY*sumY)));
	}

/**Asettaa riviosoittimen osoittamaan suhteellista rivi r vastaavaan datanhallinnan riviin.
*@see Yhteisjakaumalaskenta#annaX
*@see Yhteisjakaumalaskenta#annaY
*@see Yhteisjakaumalaskenta#siirrRiviosoitinta
*/
	public void asetaRiviosoitin(int r) { riviosoitin = annaDhRivi(r); }

/**Palauttaa x-akselilla tarkasteltavan kentn arvon riviosoittimen osoittamalla rivill.
*@see Yhteisjakaumalaskenta#asetaRiviosoitin
*@see Yhteisjakaumalaskenta#annaY
*@see Yhteisjakaumalaskenta#siirrRiviosoitinta
*/
	public float annaX() { return dh.annaArvo(riviosoitin, kenttaX, tarkenninX); }

/**Palauttaa y-akselilla tarkasteltavan kentn arvon riviosoittimen osoittamalla rivill.
*@see Yhteisjakaumalaskenta#asetaRiviosoitin
*@see Yhteisjakaumalaskenta#annaX
*@see Yhteisjakaumalaskenta#siirrRiviosoitinta
*/
	public float annaY() { return dh.annaArvo(riviosoitin, kenttaY, tarkenninY); }

/**Siirt riviosoittimen osoittamaan seuraavaan karsinnan lpisevn riviin.
*@see Yhteisjakaumalaskenta#annaX
*@see Yhteisjakaumalaskenta#annaY
*@see Yhteisjakaumalaskenta#asetaRiviosoitin
*/
	public void siirraRiviosoitinta() { riviosoitin = seuraavaRivi(riviosoitin); }
 
	protected boolean valmisLaskentaan() { return (kenttaX > -1) && (kenttaY > -1); }
	
	protected void alustaUudelleenlaskenta()
	{
		tulosumma = 0;
		summainX = new Parametrisummain();
		summainY = new Parametrisummain();
	}
	
	protected void laskeUudelleenRivi(int dhr, int r)
	{
		float x = dh.annaArvo(dhr, kenttaX, tarkenninX);
		float y = dh.annaArvo(dhr, kenttaY, tarkenninY);
		tulosumma += (double)x * (double)y;
		summainX.uusiArvo(x);
		summainY.uusiArvo(y);
	}
	
	protected void laskeLisaysrivi(int dhr, int r) { laskeUudelleenRivi(dhr, r); }
	
	private int riviosoitin;
	private int kenttaX, tarkenninX, kenttaY, tarkenninY;
	private double tulosumma;
	private Parametrisummain summainX, summainY;
}


